#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <openssl/sha.h>
#include <openssl/aes.h>

static void convertStr2Hex(const char *str, unsigned char *iv, unsigned int size)
{
    for(unsigned int i = 0; i < size; i++)
    {
        char tmpBuf[3];
        tmpBuf[0] = str[i * 2 + 0];
        tmpBuf[1] = str[i * 2 + 1];
        tmpBuf[2] = 0;

        iv[i] = strtoul(tmpBuf, 0, 16);
    }
}

int main(int argc, char *argv[])
{
    AES_KEY aesDecryptor, aesEncryptor;

    if(argc < 8)
    {
        printf("usage: ./wiiu_usb_inject PATH_TO_IMG USB_KEY PATH_TO_DATA_BLOCK META_SECTOR META_IV DATA_SECTOR DATA_IV\n");
        printf("The image is modified in place, no backup is made!\n");
        printf("Sectors, iv and usb key in hex string without 0x in front\n");
        printf("Data block file of size 0x10000 or smaller\n");
        printf("Example: ./wiiu_usb_inject wiiu_usb.img 12345678901234567890123456789012 ./haxchi/installer/data/brainage.zip 00009320 00002000923D0340003BB80000000200 00015400 00010000923DC420003BB80000000200\n");
        return 1;
    }

    if(strlen(argv[2]) != 0x20)
    {
        printf("wrong usb key size\n");
        return 1;
    }
    if(strlen(argv[4]) != 0x8)
    {
        printf("wrong meta sector size\n");
        return 1;
    }
    if(strlen(argv[5]) != 0x20)
    {
        printf("wrong meta iv size\n");
        return 1;
    }
    if(strlen(argv[6]) != 0x8)
    {
        printf("wrong data sector size\n");
        return 1;
    }
    if(strlen(argv[7]) != 0x20)
    {
        printf("wrong data iv size\n");
        return 1;
    }

	unsigned char key[0x10];
    convertStr2Hex(argv[2], key, sizeof(key));

	if (AES_set_decrypt_key(key, 128, &aesDecryptor) < 0)
    {
        printf("Failed to set AES decryption key.\n");
        return 2;
    }

	if (AES_set_encrypt_key(key, 128, &aesEncryptor) < 0)
    {
        printf("Failed to set AES encryption key.\n");
        return 2;
    }

    unsigned char iv[0x10];
    unsigned char acBuffer[0x10000];
    memset(iv, 0, sizeof(iv));

    FILE *pFile = fopen(argv[1], "r+");
    if(!pFile)
    {
        printf("Input file open failed\n");
        return 3;
    }

    int readBytes = 1;
    uint64_t totalBytes = 0;
    uint64_t offset = 0;

    uint32_t sectorData = 0;
    unsigned char ivData[0x10];
    convertStr2Hex(argv[6], (unsigned char*)&sectorData, sizeof(sectorData));
    sectorData = ((sectorData & 0xFF) << 24) | ((sectorData & 0xFF00) << 8) | ((sectorData & 0xFF0000) >> 8) | ((sectorData & 0xFF000000) >> 24);
    convertStr2Hex(argv[7], ivData, sizeof(ivData));

    fseeko(pFile, (uint64_t)sectorData << 9, SEEK_SET);

    readBytes = fread(acBuffer, 1, 0x10000, pFile);
    if(readBytes != 0x10000)
    {
        printf("read failed\n");
        fclose(pFile);
        return 3;
    }

    memcpy(iv, ivData, sizeof(iv));
    AES_cbc_encrypt(&acBuffer[0], &acBuffer[0], readBytes, &aesDecryptor, iv, AES_DECRYPT);

    unsigned char shaHashOldBlock[0x14];
    SHA1(acBuffer, readBytes, shaHashOldBlock);

    FILE *pReplace = fopen(argv[3], "rb");
    if(!pReplace)
    {
        printf("failed to open source file failed\n");
        fclose(pFile);
        return 3;
    }
    memset(acBuffer, 0 ,sizeof(acBuffer));
    readBytes = fread(acBuffer, 1, 0x10000, pReplace);
    fclose(pReplace);
    pReplace = NULL;

    if(readBytes < 0x10000)
        readBytes = 0x10000;

    unsigned char shaHashBlock[0x14];
    SHA1(acBuffer, readBytes, shaHashBlock);

    memcpy(iv, ivData, sizeof(iv));
    AES_cbc_encrypt(&acBuffer[0], &acBuffer[0], readBytes, &aesEncryptor, iv, AES_ENCRYPT);


    uint32_t sectorMeta1 = 0;
    unsigned char ivMeta1[0x10];
    unsigned char acMetaBuffer[0x2000];
    convertStr2Hex(argv[4], (unsigned char*)&sectorMeta1, sizeof(sectorMeta1));
    sectorMeta1 = ((sectorMeta1 & 0xFF) << 24) | ((sectorMeta1 & 0xFF00) << 8) | ((sectorMeta1 & 0xFF0000) >> 8) | ((sectorMeta1 & 0xFF000000) >> 24);
    convertStr2Hex(argv[5], ivMeta1, sizeof(ivMeta1));
    fseeko(pFile, (uint64_t)sectorMeta1 << 9, SEEK_SET);
    readBytes = fread(acMetaBuffer, 1, 0x2000, pFile);

    memcpy(iv, ivMeta1, sizeof(iv));
    AES_cbc_encrypt(&acMetaBuffer[0], &acMetaBuffer[0], readBytes, &aesDecryptor, iv, AES_DECRYPT);

    int found = 0;
    // since i dont wanna parse every...meh...just search the old hash
    for(int i = 0; i < readBytes; i++)
    {
        if(memcmp(&acMetaBuffer[i], shaHashOldBlock, sizeof(shaHashOldBlock)) == 0)
        {
            found = 1;
            memcpy(&acMetaBuffer[i], shaHashBlock, sizeof(shaHashBlock));
        }
    }

    if(found == 0)
    {
        printf("Hash of block not found in meta block -> do nothing\n");
        fclose(pFile);
        return 4;
    }

    fseeko(pFile, (uint64_t)sectorData << 9, SEEK_SET);

    if(fwrite(acBuffer, 1, sizeof(acBuffer), pFile) != sizeof(acBuffer))
    {
        printf("write failed\n");
        fclose(pFile);
        return 4;
    }

    unsigned char shaHash[0x14];
    memset(acMetaBuffer + 0x04, 0xFF, sizeof(shaHash));
    SHA1(acMetaBuffer, readBytes, shaHash);
    memcpy(acMetaBuffer + 0x04, shaHash, sizeof(shaHash));

    memcpy(iv, ivMeta1, sizeof(iv));
    AES_cbc_encrypt(&acMetaBuffer[0], &acMetaBuffer[0], readBytes, &aesEncryptor, iv, AES_ENCRYPT);

    fseeko(pFile, (uint64_t)sectorMeta1 << 9, SEEK_SET);
    readBytes = fwrite(acMetaBuffer, 1, readBytes, pFile);

    fclose(pFile);
    printf("Injection done\n");
    return 0;
}

